from .EpomakerCommand import EpomakerCommand
from .reports.Report import Report


class EpomakerCpuCommand(EpomakerCommand):
    """A command for setting the CPU usage on the keyboard."""

    def __init__(self, cpu: int) -> None:
        initialization_data = "22000000000000dd63007f0004000800" + f"{cpu:02x}"
        initial_report = Report(initialization_data, index=0, checksum_index=None)
        super().__init__(initial_report)



from .EpomakerCommand import EpomakerCommand
from .reports.Report import Report
from .data.constants import Profile


class EpomakerProfileCommand(EpomakerCommand):
    """A command for setting the profile on the keyboard."""

    def __init__(self, profile: Profile) -> None:
        initialization_data = (
            "07"
            f"{profile.mode.value:02x}"
            f"{profile.speed.value:02x}"
            f"{profile.brightness.value:02x}"
            f"{profile.option.value|profile.dazzle.value:02x}"
            f"{profile.rgb[0]:02x}"
            f"{profile.rgb[1]:02x}"
            f"{profile.rgb[2]:02x}"
        )
        initial_report = Report(initialization_data, index=0, checksum_index=8)
        super().__init__(initial_report)



import dataclasses
from typing import Iterator

BUFF_LENGTH: int = 128 // 2  # 128 bytes / 2 bytes per hex value


@dataclasses.dataclass()
class Report:
    """Represents the data that is sent to the keyboard."""

    header_format_string: str
    checksum_index: int | None
    index: int
    pad_on_init: bool = True
    header_format_values: dict[str, int] = dataclasses.field(default_factory=dict)
    report_bytearray: bytearray | None = None
    header_length: int | None = None

    def __post_init__(self) -> None:
        if self.header_format_values == {}:
            self.report_bytearray = bytearray.fromhex(self.header_format_string)
        else:
            self.report_bytearray = bytearray.fromhex(
                self.header_format_string.format(**self.header_format_values)
            )
        self.header_length = len(self.report_bytearray)
        if self.checksum_index is not None:
            self.report_bytearray += self._calculate_checksum(self._get_header())
        assert len(self.report_bytearray) <= BUFF_LENGTH, (
            f"Report length {len(self.report_bytearray)} exceeds the maximum length of "
            f"{BUFF_LENGTH}."
        )
        self.header_length = len(self.report_bytearray)
        if self.pad_on_init:
            self._pad()

    def _pad(self) -> None:
        """Pads the report header with zeros to the maximum length."""
        assert (
            self.report_bytearray is not None
        ), "Report bytearray must be set before padding."
        self.report_bytearray += bytes(BUFF_LENGTH - len(self.report_bytearray))

    @staticmethod
    def _calculate_checksum(buffer: bytes) -> bytes:
        sum_bits = sum(buffer) & 0xFF
        checksum = (0xFF - sum_bits) & 0xFF
        return bytes([checksum])

    def _get_checksum(self) -> bytes:
        assert (
            self.report_bytearray is not None
        ), "Report bytearray must be set before getting."
        assert (
            self.checksum_index is not None
        ), "Checksum index must be set before getting."
        return self[self.checksum_index]

    def _get_header(self) -> bytes:
        assert (
            self.header_length is not None
        ), "Header length must be set before getting."
        assert (
            self.report_bytearray is not None
        ), "Report bytearray must be set before getting."
        return self.report_bytearray[: self.header_length]

    def __getitem__(self, key: int | slice) -> bytes:
        assert (
            self.report_bytearray is not None
        ), "Report bytearray must be set before getting."
        return bytes(self.report_bytearray[key])

    def __len__(self) -> int:
        if self.report_bytearray is None:
            return 0
        return len(self.report_bytearray)

    def get_all_bytes(self) -> bytearray | None:
        assert (
            self.report_bytearray is not None
        ), "Report bytearray must be set before getting all bytes."
        return self.report_bytearray



@dataclasses.dataclass()
class ReportCollection:
    reports: list[Report] = dataclasses.field(default_factory=list)

    def __iter__(self) -> Iterator[Report]:
        for report in self.reports:
            yield report

    def __getitem__(self, key: int) -> Report:
        return self.reports[key]

    def __len__(self) -> int:
        return len(self.reports)

    def __setitem__(self, report: Report) -> None:
        assert report.index not in [
            r.index for r in self.reports
        ], f"Report index {report.index} already exists."
        assert (
            report.report_bytearray is not None
        ), "Report bytearray must be set before adding."
        self.reports.append(report)

    def append(self, report: Report) -> None:
        assert report.index not in [
            r.index for r in self.reports
        ], f"Report index {report.index} already exists."
        assert (
            report.report_bytearray is not None
        ), "Report bytearray must be set before adding."
        self.reports.append(report)
        self.reports.sort(key=lambda x: x.index)

    def iter_report_bytes(self) -> Iterator[bytes]:
        for report in self.reports:
            assert (
                report.report_bytearray is not None
            ), "Report bytearray must be set before iterating."
            yield report.report_bytearray



import dataclasses
from typing import Iterator
from .reports.Report import Report, ReportCollection
import numpy as np
import numpy.typing as npt


@dataclasses.dataclass(frozen=True)
class CommandStructure:
    number_of_starter_reports: int = 1
    number_of_data_reports: int = 0
    number_of_footer_reports: int = 0

    def __len__(self) -> int:
        return (
            self.number_of_starter_reports
            + self.number_of_data_reports
            + self.number_of_footer_reports
        )


class EpomakerCommand:
    """A command is basically just a wrapper around a numpy array of bytes.

    The command must have x dimension of 128 bytes and be padded with zeros if the command is
    less than 128 bytes.

    The command may have a y dimension of at least 1, and more if the command is a series of
    packets to be sent (eg image, key colours). Each row of extra packets must also have an
    associated header according to the commannd type.
    """

    def __init__(
        self,
        initial_report: Report,
        structure: CommandStructure = CommandStructure(),
    ) -> None:
        self.reports: ReportCollection = ReportCollection()
        self.structure: CommandStructure = structure
        # If there are data reports, the command must be prepared before sending
        self.report_data_prepared: bool = structure.number_of_data_reports == 0
        self.report_footer_prepared: bool = structure.number_of_footer_reports == 0
        self._insert_report(initial_report)

    def _insert_report(self, report: Report) -> None:
        assert report.index < len(self.structure), (
            f"Report index {report.index} exceeds the number of reports "
            f"{len(self.structure)}."
        )
        # assert report.index not in [r.index for r in self.reports], (
        #     f"Report index {report.index} already exists."
        #     )
        self.reports.append(report)

    @staticmethod
    def _np16_to_np8(data_16bit: npt.NDArray[np.uint16]) -> npt.NDArray[np.uint8]:
        """Converts a numpy array of 16-bit numbers to 8-bit numbers."""
        new_shape = (data_16bit.shape[0], data_16bit.shape[1] * 2)
        data_8bit_flat = np.empty(data_16bit.size * 2, dtype=np.uint8)

        data_8bit_flat[0::2] = (data_16bit >> 8).flatten()  # High bytes
        data_8bit_flat[1::2] = (data_16bit & 0xFF).flatten()  # Low bytes

        return data_8bit_flat.reshape(new_shape)

    def __iter__(self) -> Iterator[Report]:
        for report in self.reports:
            yield report

    def __getitem__(self, key: int) -> Report:
        report = next((r for r in self.reports if r.index == key), None)
        assert report is not None, f"Report {key} not found."
        return report

    def iter_report_bytes(self) -> Iterator[bytes]:
        for report_bytes in self.reports.iter_report_bytes():
            yield report_bytes



from enum import Enum
from dataclasses import dataclass
from typing import Iterator

BUFF_LENGTH = 128 // 2  # 128 bytes / 2 bytes per int
IMAGE_DIMENSIONS = (162, 173)


@dataclass(frozen=True)
class Profile:
    """Thanks to https://gitlab.com/CalcProgrammer1/OpenRGB/-/blob/master/Controllers/EpomakerController/EpomakerController.h"""

    class Mode(Enum):
        ALWAYS_ON = 0x01
        DYNAMIC_BREATHING = 0x02
        SPECTRUM_CYCLE = 0x03
        DRIFT = 0x04
        WAVES_RIPPLE = 0x05
        STARS_TWINKLE = 0x06
        STEADY_STREAM = 0x07
        SHADOWING = 0x08
        PEAKS_RISING_ONE_AFTER_ANOTHER = 0x09
        SINE_WAVE = 0x0A
        CAISPRING_SURGING = 0x0B
        FLOWERS_BLOOMING = 0x0C
        LASER = 0x0E
        PEAK_TURN = 0x0F
        INCLINED_RAIN = 0x10
        SNOW = 0x11
        METEOR = 0x12
        THROUGH_THE_SNOW_NON_TRACE = 0x13
        LIGHT_SHADOW = 0x15

    class Speed(Enum):
        MIN = 0x00
        MAX = 0x05
        MAX_SPECIAL = 0x04
        DEFAULT = 0x04

    class Brightness(Enum):
        MIN = 0x00
        MAX = 0x04
        DEFAULT = 0x04

    class Dazzle(Enum):
        OFF = 0x07
        ON = 0x08

    class Option(Enum):
        OFF = 0x00
        ON = 0x01
        DEFAULT = 0x00
        DRIFT_RIGHT = 0x00
        DRIFT_LEFT = 0x10
        DRIFT_DOWN = 0x20
        DRIFT_UP = 0x30
        STEADY_STREAM_ZIG_ZAG = 0x00
        STEADY_STREAM_RETURN = 0x10
        CAISPRING_SURGING_OUT = 0x00
        CAISPRING_SURGING_IN = 0x10
        FLOWERS_BLOOMING_RIGHT = 0x00
        FLOWERS_BLOOMING_LEFT = 0x10
        PEAK_TURN_ANTI_CLOCKWISE = 0x00
        PEAK_TURN_CLOCKWISE = 0x10

    mode: Mode
    speed: Speed
    brightness: Brightness
    dazzle: Dazzle
    option: Option
    rgb: tuple[int, int, int]


@dataclass(frozen=True)
class KeyboardKey(Enum):
    ESC = 0
    BACKQUOTE = 1
    TAB = 2
    CAPS = 3
    LEFT_SHIFT = 4
    LEFT_CTRL = 5
    F1 = 6
    NUMROW_1 = 7
    Q = 8
    A = 9
    BACKSLASH = 10
    DIAL = 11
    F2 = 12
    NUMROW_2 = 13
    W = 14
    S = 15
    Z = 16
    LEFT_WIN = 17
    F3 = 18
    NUMROW_3 = 19
    E = 20
    D = 21
    X = 22
    LEFT_ALT = 23
    F4 = 24
    NUMROW_4 = 25
    R = 26
    F = 27
    C = 28
    F5 = 30
    NUMROW_5 = 31
    T = 32
    G = 33
    V = 34
    F6 = 36
    NUMROW_6 = 37
    Y = 38
    H = 39
    B = 40
    SPACE = 41
    F7 = 42
    NUMROW_7 = 43
    U = 44
    J = 45
    N = 46
    F8 = 48
    NUMROW_8 = 49
    I = 50
    K = 51
    M = 52
    RIGHT_ALT = 53
    F9 = 54
    NUMROW_9 = 55
    O = 56
    L = 57
    COMMA = 58
    FN = 59
    F10 = 60
    NUMROW_0 = 61
    P = 62
    SEMICOLON = 63
    DOT = 64
    RIGHT_CTRL = 65
    F11 = 66
    NUMROW_MINUS = 67
    LEFT_BRACKET = 68
    QUOTE = 69
    SLASH = 70
    LEFT = 71
    F12 = 72
    NUMROW_EQUAL = 73
    RIGHT_BRACKET = 74
    HASH = 75
    RIGHT_SHIFT = 76
    DOWN = 77
    DEL = 78
    BACKSPACE = 79
    ENTER = 80
    NUMPAD_4 = 81
    UP = 82
    RIGHT = 83
    PGUP = 84
    NUMLOCK = 85
    NUMPAD_7 = 86
    NUMPAD_5 = 87
    NUMPAD_1 = 88
    NUMPAD_0 = 89
    PGDOWN = 90
    NUMPAD_SLASH = 91
    NUMPAD_8 = 92
    NUMPAD_6 = 93
    NUMPAD_2 = 94
    NUMPAD_DOT = 95
    NUMPAD_MINUS = 96
    NUMPAD_ASTERISK = 97
    NUMPAD_9 = 98
    NUMPAD_PLUS = 99
    NUMPAD_3 = 100
    NUMPAD_ENTER = 101


ALL_KEYBOARD_KEYS = [KeyboardKey[e.name] for e in KeyboardKey]



from dataclasses import dataclass

from .EpomakerCommand import EpomakerCommand, CommandStructure
from .reports.Report import Report, BUFF_LENGTH
from .reports.ReportWithData import ReportWithData
from .data.constants import KeyboardKey
from typing import Iterator


class KeyMap:
    """Map a KeyboardKey index to an RGB value."""

    def __init__(self) -> None:
        self.key_map: dict[int, tuple[int, int, int]] = {}

    def __getitem__(self, key: KeyboardKey) -> tuple[int, int, int]:
        return self.key_map[key.value]

    def __setitem__(self, key: KeyboardKey, value: tuple[int, int, int]) -> None:
        self.key_map[key.value] = value

    def __iter__(self) -> Iterator[tuple[int, tuple[int, int, int]]]:
        return iter(self.key_map.items())


@dataclass(frozen=True)
class KeyboardRGBFrame:
    """A keyboard frame consists of a map of keys to RGB values as well as a time in milliseconds
    to display the frame."""

    key_map: KeyMap
    time_ms: int
    length: int = 7
    index: int = 0


class EpomakerKeyRGBCommand(EpomakerCommand):
    """Change a selection of keys to specific RGB values."""

    def __init__(self, frames: list[KeyboardRGBFrame]) -> None:
        initialization_data = "18000000000000e7"
        self.report_data_header_length = 8
        data_reports_per_frame = 7
        structure = CommandStructure(
            number_of_starter_reports=1,
            number_of_data_reports=len(frames) * data_reports_per_frame,
            number_of_footer_reports=0,
        )
        initial_report = Report(initialization_data, index=0, checksum_index=None)
        super().__init__(initial_report, structure)

        report_index = 1
        data_buffer_length = BUFF_LENGTH - self.report_data_header_length
        for frame in frames:
            for this_frame_report_index in range(0, data_reports_per_frame):
                report = ReportWithData(
                    "19{this_frame_report_index:02x}{frame_index:02x}{total_frames:02x}{frame_time:02x}0000",
                    index=report_index,
                    header_format_values={
                        "this_frame_report_index": this_frame_report_index,
                        "frame_index": frame.index,
                        "total_frames": len(frames),
                        "frame_time": frame.time_ms,
                    },
                    checksum_index=7,
                )
                # Zero out the data buffer
                data_byterarray = bytearray(data_buffer_length)
                for key_index, rgb in frame.key_map:
                    # For each key, set the RGB values in the data buffer
                    for i, colour in enumerate(rgb):
                        # R, G, B individually
                        this_frame_colour_index = (
                            (key_index * 3)
                            - (this_frame_report_index * data_buffer_length)
                            + i
                        )
                        if 0 <= this_frame_colour_index < len(data_byterarray):
                            data_byterarray[this_frame_colour_index] = colour
                report.add_data(data_byterarray)
                self._insert_report(report)
                report_index += 1

    def get_data_reports(self) -> list[ReportWithData]:
        return [r for r in self.reports if isinstance(r, ReportWithData)]

    def report_data_contain_index(self, report: ReportWithData, index: int) -> bool:
        """Checks if the provided report contains the specified index if all the data portions of
        the reports were to be indexed linearly.

        Uses BUFF_LENGTH - self.report_data_header_length so this function can be used before
        the data is set.
        """
        report_index_count = 0
        data_buffer_length = BUFF_LENGTH - self.report_data_header_length
        for report in self.get_data_reports():
            report_data = report[self.report_data_header_length :]
            if report_index_count <= index < report_index_count + data_buffer_length:
                return True
            report_index_count += len(report_data)
        return False



import dataclasses

from .Report import Report


@dataclasses.dataclass()
class ReportWithData(Report):
    def __init__(
        self,
        header_format_string: str,
        checksum_index: int | None,
        index: int,
        header_format_values: dict[str, int] = dataclasses.field(default_factory=dict),
        report_bytearray: bytearray | None = None,
        header_length: int | None = None,
        report_data: bytearray | None = None,
        prepared: bool = False,
    ) -> None:
        super().__init__(
            header_format_string,
            checksum_index,
            index,
            False,
            header_format_values,
            report_bytearray,
            header_length,
        )
        self.report_data: bytearray | None = report_data
        self.prepared: bool = prepared
        if self.report_data is not None:
            self.prepared = True

    def add_data(self, data: bytes) -> None:
        assert self.prepared is False, "Report data has already been set."
        assert (
            self.report_bytearray is not None
        ), "Report bytearray must be set before adding data."
        self.report_data = bytearray(data)
        self.report_bytearray += self.report_data
        self._pad()
        self.prepared = True

    # def __getitem__(self, key: int | slice) -> bytes:
    #     assert self.report_bytearray is not None, "Report bytearray must be set before getting."
    #     assert self.report_data is not None, "Report data must be set before getting."
    #     return bytes((self.report_bytearray + self.report_data)[key])



import cv2
import numpy as np

from .EpomakerCommand import EpomakerCommand, CommandStructure
from .data.constants import IMAGE_DIMENSIONS
from .reports.Report import Report, BUFF_LENGTH
from .reports.ReportWithData import ReportWithData


class EpomakerImageCommand(EpomakerCommand):
    """A command for sending images to the keyboard."""

    def __init__(self) -> None:
        initialization_data = "a5000100f4da008b0000a2ad"
        self.report_data_header_length = 8
        structure = CommandStructure(
            number_of_starter_reports=1,
            number_of_data_reports=1000,
            number_of_footer_reports=1,
        )
        initial_report = Report(initialization_data, index=0, checksum_index=None)
        super().__init__(initial_report, structure)

    def get_data_reports(self) -> list[ReportWithData]:
        return [r for r in self.reports if isinstance(r, ReportWithData)]

    @staticmethod
    def _encode_rgb565(r: int, g: int, b: int) -> int:
        # Mask the bits of each color to fit the 5-6-5 format
        r_565 = (r & 0b11111000) << 8  # Red: 5 bits
        g_565 = (g & 0b11111100) << 3  # Green: 6 bits
        b_565 = (b & 0b11111000) >> 3  # Blue: 5 bits

        # Combine the channels into one 16-bit number
        rgb_565 = r_565 | g_565 | b_565

        return rgb_565

    @staticmethod
    def _decode_rgb565(pixel: int) -> tuple[int, int, int]:
        r = (pixel & 0xF800) >> 8  # Red: 5 bits
        g = (pixel & 0x07E0) >> 3  # Green: 6 bits
        b = (pixel & 0x001F) << 3  # Blue: 5 bits

        # We need to adjust them because we're expanding to 8 bits per channel
        r |= r >> 5
        g |= g >> 6
        b |= b >> 5

        return (r, g, b)

    def encode_image(self, image_path: str) -> None:
        """
        Encode an image to 16-bit RGB565 according to IMAGE_DIMENSIONS and accounting for packet
        headers.

        The image is also rotated and flipped since this seems to be what the keyboard is expecting.
        """
        image = cv2.imread(image_path)
        image = cv2.resize(image, IMAGE_DIMENSIONS)
        image = cv2.flip(image, 0)
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        image_16bit = np.zeros(
            (IMAGE_DIMENSIONS[0], IMAGE_DIMENSIONS[1]), dtype=np.uint16
        )
        try:
            for y in range(image.shape[0]):
                for x in range(image.shape[1]):
                    r, g, b = image[y, x]
                    image_16bit[y, x] = self._encode_rgb565(r, g, b)
        except Exception as e:
            print(f"Exception while converting image: {e}")

        image_8bit_flattened = np.ndarray.flatten(self._np16_to_np8(image_16bit))
        data_buff_length = BUFF_LENGTH - self.report_data_header_length
        data_buff_pointer = 0
        for report_index in range(0, self.structure.number_of_data_reports):
            report_index_bytes = report_index.to_bytes(2, "big")
            report = ReportWithData(
                header_format_string="25000100{report_index_bytes_upper:02x}{report_index_bytes_lower:02x}38",
                index=report_index + self.structure.number_of_starter_reports,
                header_format_values={
                    "report_index_bytes_upper": report_index_bytes[1],
                    "report_index_bytes_lower": report_index_bytes[0],
                },
                checksum_index=7,
            )
            report.add_data(
                image_8bit_flattened[
                    data_buff_pointer : data_buff_pointer + data_buff_length
                ].tobytes()
            )
            data_buff_pointer += data_buff_length
            self._insert_report(report)

        assert len(self.get_data_reports()) == self.structure.number_of_data_reports, (
            f"Expected {self.structure.number_of_data_reports} reports, got "
            f"{len(self.get_data_reports())}."
        )

        self.report_data_prepared = True

        # Add the footer report
        footer_index_bytes = (
            self.structure.number_of_starter_reports
            + self.structure.number_of_data_reports
            - self.structure.number_of_footer_reports
        ).to_bytes(2, "big")
        footer_report = ReportWithData(
            header_format_string="25000100{footer_index_bytes_upper:02x}{footer_index_bytes_lower:02x}34",
            index=self.structure.number_of_starter_reports
            + self.structure.number_of_data_reports,
            header_format_values={
                "footer_index_bytes_upper": footer_index_bytes[1],
                "footer_index_bytes_lower": footer_index_bytes[0],
            },
            checksum_index=7,
        )
        footer_report.add_data(
            image_8bit_flattened[
                data_buff_pointer : data_buff_pointer + data_buff_length
            ].tobytes()
        )
        # Need some padding at the end of the image data
        footer_report._pad()
        self._insert_report(footer_report)

        self.report_footer_prepared = True

        assert len(self.reports) == len(
            self.structure
        ), f"Expected {len(self.structure)} reports, got {len(self.reports)}."



# src/epomakercontroller/cli.py

import time
import click
from epomakercontroller.commands import (
    EpomakerKeyRGBCommand,
    EpomakerProfileCommand,
)
from epomakercontroller.commands.data.constants import ALL_KEYBOARD_KEYS, Profile
from epomakercontroller import EpomakerController
import psutil

@click.group()
def cli() -> None:
    """EpomakerController CLI."""
    pass

@cli.command()
@click.argument('image_path', type=click.Path(exists=True))
def upload_image(image_path: str) -> None:
    """Upload an image to the Epomaker device."""
    try:
        controller = EpomakerController(dry_run=False)
        if controller.open_device():
            print("Uploading, you should see the status on the keyboard screen")
            controller.send_image(image_path)
            click.echo("Image uploaded successfully.")
        controller.close_device()
    except Exception as e:
        click.echo(f"Failed to upload image: {e}")

@cli.command()
@click.argument('r', type=int)
@click.argument('g', type=int)
@click.argument('b', type=int)
def set_rgb_all_keys(r: int, g: int, b: int)  -> None:
    """Set RGB colour for all keys."""
    try:
        mapping = EpomakerKeyRGBCommand.KeyMap()
        for key in ALL_KEYBOARD_KEYS:
            mapping[key] = (r, g, b)
        frames = [EpomakerKeyRGBCommand.KeyboardRGBFrame(mapping, 50)]
        controller = EpomakerController(dry_run=False)
        if controller.open_device():
            controller.send_keys(frames)
            click.echo(f"All keys set to RGB({r}, {g}, {b}) successfully.")
        controller.close_device()
    except Exception as e:
        click.echo(f"Failed to set RGB for all keys: {e}")

@cli.command()
def cycle_light_modes() -> None:
    """Cycle through the light modes."""
    try:
        controller = EpomakerController(dry_run=False)
        if not controller.open_device():
            click.echo("Failed to open device.")
            return

        print(f"Cycling through {len(Profile.Mode)} modes, waiting 5 seconds on each")
        counter = 1
        for mode in Profile.Mode:
            profile = Profile(
                mode=mode,
                speed=Profile.Speed.DEFAULT,
                brightness=Profile.Brightness.DEFAULT,
                dazzle=Profile.Dazzle.OFF,
                option=Profile.Option.OFF,
                rgb=(180, 180, 180),
            )
            command = EpomakerProfileCommand.EpomakerProfileCommand(profile)
            controller._send_command(command)
            click.echo(f"[{counter}/{len(Profile.Mode)}] Cycled to light mode: {mode.name}")
            time.sleep(5)
            counter += 1

        controller.close_device()
        click.echo("Cycled through all light modes successfully.")
    except Exception as e:
        click.echo(f"Failed to cycle light modes: {e}")

@cli.command()
def send_time() -> None:
    """Send the current time to the Epomaker device."""
    try:
        controller = EpomakerController(dry_run=False)
        if controller.open_device():
            controller.send_time()
            click.echo("Time sent successfully.")
        controller.close_device()
    except Exception as e:
        click.echo(f"Failed to send time: {e}")

@cli.command()
@click.argument('temperature', type=int)
def send_temperature(temperature: int) -> None:
    """Send temperature to the Epomaker screen."""
    try:
        controller = EpomakerController(dry_run=False)
        if controller.open_device():
            controller.send_temperature(temperature)
            click.echo("Temperature sent successfully.")
        controller.close_device()
    except Exception as e:
        click.echo(f"Failed to send temperature: {e}")

@cli.command()
@click.argument('cpu', type=int)
def send_cpu(cpu: int) -> None:
    """Send CPU usage percentage to the Epomaker screen."""
    try:
        controller = EpomakerController(dry_run=False)
        if controller.open_device():
            controller.send_cpu(cpu)
            click.echo("CPU usage sent successfully.")
        controller.close_device()
    except Exception as e:
        click.echo(f"Failed to send CPU usage: {e}")

@cli.command()
@click.argument('temp_key', type=str, required=False)
def start_daemon(temp_key: str | None) -> None:
    """Start a daemon to update the CPU usage and optionally a temperature."""
    try:
        controller = EpomakerController(dry_run=False)
        while True:
            if not controller.open_device():
                click.echo("Failed to open device.")
                return
            # Set current time and date
            controller.send_time()

            # Get CPU usage
            cpu_usage = round(psutil.cpu_percent(interval=1))
            click.echo(f"CPU Usage: {cpu_usage}%")
            controller.send_cpu(int(cpu_usage), from_daemon=True)

            # Get device temperature using the provided key
            if temp_key:
                try:
                    temps = psutil.sensors_temperatures()
                    if temp_key in temps:
                        cpu_temp = temps[temp_key][0].current
                        rounded_cpu_temp = round(cpu_temp)
                        click.echo(f"CPU Temperature ({temp_key}): {rounded_cpu_temp}°C")
                        controller.send_temperature(rounded_cpu_temp)
                    else:
                        available_keys = list(temps.keys())
                        click.echo(f"Temperature key '{temp_key}' not found. Available keys: {available_keys}")
                except AttributeError:
                    click.echo("Temperature monitoring not supported on this system.")


            controller.close_device()

            time.sleep(3)
    except KeyboardInterrupt:
        click.echo("Daemon interrupted by user.")
    except Exception as e:
        click.echo(f"Failed to start daemon: {e}")


@cli.command()
def list_temp_devices() -> None:
    """List available temperature devices."""
    try:
        temps = psutil.sensors_temperatures()
        if not temps:
            click.echo("No temperature sensors found.")
            return

        for key, entries in temps.items():
            click.echo(f"\nTemperature key: {key}")
            for entry in entries:
                click.echo(f"  Label: {entry.label or 'N/A'}")
                click.echo(f"  Current: {entry.current}°C")
                click.echo(f"  High: {entry.high or 'N/A'}°C")
                click.echo(f"  Critical: {entry.critical or 'N/A'}°C")
    except AttributeError:
        click.echo("Temperature monitoring not supported on this system.")


if __name__ == "__main__":
    cli()



from datetime import datetime
import time
import hid

from .commands import (
    EpomakerCommand,
    EpomakerImageCommand,
    EpomakerTimeCommand,
    EpomakerTempCommand,
    EpomakerCpuCommand,
    EpomakerKeyRGBCommand,
)
from .commands.data.constants import BUFF_LENGTH

VENDOR_ID = 0x3151
PRODUCT_ID = 0x4010
INTERFACE_NUMBER = 1

class EpomakerController:
    """
    EpomakerController class represents a controller for an Epomaker USB HID device.

    Attributes:
        vendor_id (int): The vendor ID of the USB HID device.
        product_id (int): The product ID of the USB HID device.
        device (hid.device): The HID device object.

    Methods:
        open_device(vendor_id, product_id): Opens the USB HID device and prints device information.
        send_basic_command(command, command_type): Sends a command to the HID device.
        close_device(): Closes the USB HID device.
        format_current_time(): Gets the current time and formats it into the required byte string
        format.
    """

    def __init__(
        self,
        vendor_id: int = VENDOR_ID,
        product_id: int = PRODUCT_ID,
        dry_run: bool = True,
    ) -> None:
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.device = hid.device()
        self.dry_run = dry_run
        print("WARNING: If this program errors out or you cancel early, the keyboard may become unresponsive. It should work fine again if you unplug and plug it back in!")


    def open_device(self) -> bool:
        """
        Opens the USB HID device and prints device information.
        Returns True if the device is successfully opened.
        """
        if self.dry_run:
            print("Dry run: skipping device open")
            return True
        try:
            # Find the device with the specified interface number so we can open by path
            # This way we don't block usage of the keyboard whilst the device is open
            device_list = hid.enumerate(self.vendor_id, self.product_id)
            device_path = None
            for device in device_list:
                if device['interface_number'] == INTERFACE_NUMBER:
                    device_path = device['path']
                    break

            if device_path is None:
                raise ValueError(f"No device found with interface number {INTERFACE_NUMBER}")

            self.device = hid.device()
            self.device.open_path(device_path)
            # print(f"Manufacturer: {self.device.get_manufacturer_string()}")
            # print(f"Product: {self.device.get_product_string()}")
            # print(f"Serial No: {self.device.get_serial_number_string()}")
            return True
        except Exception as e:
            print(f"Failed to open device: {e}")
            return False

    def _send_command(
        self, command: EpomakerCommand.EpomakerCommand, sleep_time: float = 0.1
    ) -> None:
        """
        Sends a command to the HID device.
        """
        assert command.report_data_prepared
        for packet in command:
            assert len(packet) == BUFF_LENGTH
            if self.dry_run:
                print(f"Dry run: skipping command send: {packet!r}")
            elif self.device:
                self.device.send_feature_report(packet.get_all_bytes())
            time.sleep(sleep_time)

    @staticmethod
    def _assert_range(value: int, r: range = range(0, 100)) -> bool:
        return value in r

    def send_image(self, image_path: str) -> None:
        """
        Sends an image to the HID device.
        """
        image_command = EpomakerImageCommand.EpomakerImageCommand()
        image_command.encode_image(image_path)
        self._send_command(image_command, sleep_time=0.01)

    def send_time(self, time: datetime = datetime.now()) -> None:
        """
        Sends the current time to the HID device.
        """
        time_command = EpomakerTimeCommand.EpomakerTimeCommand(time)
        self._send_command(time_command)

    def send_temperature(self, temperature: int) -> None:
        """
        Sends the temperature to the HID device.
        """
        if not self._assert_range(temperature):
            raise ValueError("Temperature must be in range 0-100")
        temperature_command = EpomakerTempCommand.EpomakerTempCommand(temperature)
        self._send_command(temperature_command)

    def send_cpu(self, cpu: int, from_daemon: bool=False) -> None:
        """
        Sends the CPU percentage to the HID device.
        """
        if not from_daemon:
            if not self._assert_range(cpu):
                raise ValueError("CPU percentage must be in range 0-100")
        cpu_command = EpomakerCpuCommand.EpomakerCpuCommand(cpu)
        self._send_command(cpu_command)

    def send_keys(self, frames: list[EpomakerKeyRGBCommand.KeyboardRGBFrame]) -> None:
        rgb_command = EpomakerKeyRGBCommand.EpomakerKeyRGBCommand(frames)
        self._send_command(rgb_command)

    def close_device(self) -> None:
        """
        Closes the USB HID device.
        """
        if self.device:
            self.device.close()
            self.device = None

    def __del__(self) -> None:
        self.close_device()



from .EpomakerCommand import EpomakerCommand
from .reports.Report import Report


class EpomakerTempCommand(EpomakerCommand):
    """A command for setting the temperature on the keyboard."""

    def __init__(self, temp: int) -> None:
        initialization_data = "2a000000000000d5" + f"{temp:02x}"
        initial_report = Report(initialization_data, index=0, checksum_index=None)
        super().__init__(initial_report)



from datetime import datetime
from .EpomakerCommand import EpomakerCommand
from .reports.Report import Report


class EpomakerTimeCommand(EpomakerCommand):
    """A command for setting the time on the keyboard."""

    def __init__(self, time: datetime) -> None:
        initialization_data = "28000000000000d7" + self._format_time(time)
        initial_report = Report(initialization_data, index=0, checksum_index=None)
        super().__init__(initial_report)

    @staticmethod
    def _format_time(time: datetime) -> str:
        """
        Gets the current time and formats it into the required byte string format.
        Returns the formatted command string to send to the device.
        """
        print("Using:", time)

        # Example of formatting for a specific date and time format
        # Adjust the formatting based on your specific requirements
        year = time.year
        month = time.month
        day = time.day
        hour = time.hour
        minute = time.minute
        second = time.second

        # Convert to hexadecimal strings
        year_hex = f"{year:04x}"
        month_hex = f"{month:02x}"
        day_hex = f"{day:02x}"
        hour_hex = f"{hour:02x}"
        minute_hex = f"{minute:02x}"
        second_hex = f"{second:02x}"

        command = f"{year_hex}{month_hex}{day_hex}{hour_hex}{minute_hex}{second_hex}"

        return command
